use crate::ast::node::generics::GenericParameters;
use crate::ast::node::implements_list::ImplementsList;
use crate::ast::node::level::ClassLevelDeclaration;
use crate::ast::node::names::Identifier;
use crate::ast::node::type_use::TypeUse;

#[derive(Debug)]
pub struct ClassDeclaration {
    is_public: bool,
    identifier: Box<Identifier>,
    generics: Box<GenericParameters>,
    class_constructor: Option<Box<ClassConstructor>>,
    implements: Box<ImplementsList>,
    declarations: Vec<Box<ClassLevelDeclaration>>,
}

impl ClassDeclaration {
    pub fn new(
        is_public: bool,
        identifier: Box<Identifier>,
        generics: Box<GenericParameters>,
        class_constructor: Option<Box<ClassConstructor>>,
        implements: Box<ImplementsList>,
        declarations: Vec<Box<ClassLevelDeclaration>>,
    ) -> Self {
        Self {
            is_public,
            identifier,
            generics,
            class_constructor,
            implements,
            declarations,
        }
    }

    pub fn is_public(&self) -> bool {
        self.is_public
    }

    pub fn identifier(&self) -> &Identifier {
        &self.identifier
    }

    pub fn identifier_mut(&mut self) -> &mut Identifier {
        &mut self.identifier
    }

    pub fn generics(&self) -> &GenericParameters {
        &self.generics
    }

    pub fn generics_mut(&mut self) -> &mut GenericParameters {
        &mut self.generics
    }

    pub fn class_constructor(&self) -> Option<&ClassConstructor> {
        if let Some(class_constructor) = &self.class_constructor {
            Some(class_constructor.as_ref())
        } else {
            None
        }
    }

    pub fn class_constructor_mut(&mut self) -> Option<&mut ClassConstructor> {
        if let Some(class_constructor) = &mut self.class_constructor {
            Some(class_constructor.as_mut())
        } else {
            None
        }
    }

    pub fn implements(&self) -> &ImplementsList {
        &self.implements
    }

    pub fn implements_mut(&mut self) -> &mut ImplementsList {
        &mut self.implements
    }

    pub fn declarations(&self) -> Vec<&ClassLevelDeclaration> {
        self.declarations.iter().map(Box::as_ref).collect()
    }

    pub fn declarations_mut(&mut self) -> Vec<&mut ClassLevelDeclaration> {
        self.declarations.iter_mut().map(Box::as_mut).collect()
    }
}

#[derive(Debug)]
pub struct ClassConstructor(Vec<Box<ClassConstructorParameter>>);

impl ClassConstructor {
    pub fn new(parameters: Vec<Box<ClassConstructorParameter>>) -> Self {
        Self(parameters)
    }

    pub fn parameters(&self) -> Vec<&ClassConstructorParameter> {
        self.0.iter().map(|p| p.as_ref()).collect()
    }

    pub fn parameters_mut(&mut self) -> Vec<&mut ClassConstructorParameter> {
        self.0.iter_mut().map(|p| p.as_mut()).collect()
    }
}

#[derive(Debug)]
pub enum ClassConstructorParameter {
    Property(Box<PropertyDeclaration>),
    Field(Box<FieldDeclaration>),
}

#[derive(Debug)]
pub struct PropertyDeclaration {
    is_mutable: bool,
    identifier: Box<Identifier>,
    declared_type: Box<TypeUse>,
}

impl PropertyDeclaration {
    pub fn new(is_mutable: bool, identifier: Box<Identifier>, declared_type: Box<TypeUse>) -> Self {
        Self {
            is_mutable,
            identifier,
            declared_type,
        }
    }

    pub fn is_mutable(&self) -> bool {
        self.is_mutable
    }

    pub fn identifier(&self) -> &Identifier {
        &self.identifier
    }

    pub fn identifier_mut(&mut self) -> &mut Identifier {
        &mut self.identifier
    }

    pub fn declared_type(&self) -> &TypeUse {
        &self.declared_type
    }

    pub fn declared_type_mut(&mut self) -> &mut TypeUse {
        &mut self.declared_type
    }
}

#[derive(Debug)]
pub struct FieldDeclaration {
    is_mutable: bool,
    identifier: Box<Identifier>,
    declared_type: Box<TypeUse>,
}

impl FieldDeclaration {
    pub fn new(is_mutable: bool, identifier: Box<Identifier>, declared_type: Box<TypeUse>) -> Self {
        Self {
            is_mutable,
            identifier,
            declared_type,
        }
    }

    pub fn is_mutable(&self) -> bool {
        self.is_mutable
    }

    pub fn identifier(&self) -> &Identifier {
        &self.identifier
    }

    pub fn identifier_mut(&mut self) -> &mut Identifier {
        &mut self.identifier
    }

    pub fn declared_type(&self) -> &TypeUse {
        &self.declared_type
    }

    pub fn declared_type_mut(&mut self) -> &mut TypeUse {
        &mut self.declared_type
    }
}
