use crate::ast::node::function::{FunctionTypeModifier, Parameters, ReturnType};
use crate::ast::node::generics::{GenericArguments, GenericParameters};
use crate::ast::node::names::FullyQualifiedName;
use crate::ast::node::tuple_arguments::TupleArguments;

#[derive(Debug)]
pub enum TypeUse {
    Primitive(Box<PrimitiveTypeUse>),
    InterfaceOrClass(Box<InterfaceOrClassTypeUse>),
    Tuple(Box<TupleTypeUse>),
    Function(Box<FunctionTypeUse>),
}

#[derive(Debug)]
pub enum PrimitiveTypeUse {
    Byte,
    Short,
    Char,
    Int,
    Long,
    Double,
    Bool,
    String,
    Array(Option<Box<GenericArguments>>),
    Any,
    Void,
}

#[derive(Debug)]
pub struct InterfaceOrClassTypeUse {
    borrow_count: usize,
    is_mutable: bool,
    fqn: Box<FullyQualifiedName>,
    generics: Box<GenericArguments>,
}

impl InterfaceOrClassTypeUse {
    pub fn new(
        borrow_count: usize,
        is_mutable: bool,
        fqn: Box<FullyQualifiedName>,
        generics: Box<GenericArguments>,
    ) -> Self {
        Self {
            borrow_count,
            is_mutable,
            fqn,
            generics,
        }
    }

    pub fn borrow_count(&self) -> usize {
        self.borrow_count
    }

    pub fn is_mutable(&self) -> bool {
        self.is_mutable
    }

    pub fn fqn(&self) -> &FullyQualifiedName {
        &self.fqn
    }

    pub fn fqn_mut(&mut self) -> &mut FullyQualifiedName {
        &mut self.fqn
    }

    pub fn generics(&self) -> &GenericArguments {
        &self.generics
    }

    pub fn generics_mut(&mut self) -> &mut GenericArguments {
        &mut self.generics
    }
}

#[derive(Debug)]
pub struct TupleTypeUse {
    borrow_count: usize,
    is_mutable: bool,
    arguments: Box<TupleArguments>,
}

impl TupleTypeUse {
    pub fn new(borrow_count: usize, is_mutable: bool, arguments: Box<TupleArguments>) -> Self {
        Self {
            borrow_count,
            is_mutable,
            arguments,
        }
    }

    pub fn borrow_count(&self) -> usize {
        self.borrow_count
    }

    pub fn is_mutable(&self) -> bool {
        self.is_mutable
    }

    pub fn arguments(&self) -> &TupleArguments {
        &self.arguments
    }

    pub fn arguments_mut(&mut self) -> &mut TupleArguments {
        &mut self.arguments
    }
}

#[derive(Debug)]
pub struct FunctionTypeUse {
    borrow_count: usize,
    function_modifier: Option<FunctionTypeModifier>,
    generics: Box<GenericParameters>,
    parameters: Box<Parameters>,
    return_type: Box<ReturnType>,
}

impl FunctionTypeUse {
    pub fn new(
        borrow_count: usize,
        function_modifier: Option<FunctionTypeModifier>,
        generics: Box<GenericParameters>,
        parameters: Box<Parameters>,
        return_type: Box<ReturnType>,
    ) -> Self {
        Self {
            borrow_count,
            function_modifier,
            generics,
            parameters,
            return_type,
        }
    }

    pub fn borrow_count(&self) -> usize {
        self.borrow_count
    }

    pub fn function_modifier(&self) -> Option<&FunctionTypeModifier> {
        self.function_modifier.as_ref()
    }

    pub fn generics(&self) -> &GenericParameters {
        &self.generics
    }

    pub fn generics_mut(&mut self) -> &mut GenericParameters {
        &mut self.generics
    }

    pub fn parameters(&self) -> &Box<Parameters> {
        &self.parameters
    }

    pub fn parameters_mut(&mut self) -> &mut Box<Parameters> {
        &mut self.parameters
    }

    pub fn return_type(&self) -> &ReturnType {
        &self.return_type
    }

    pub fn return_type_mut(&mut self) -> &mut ReturnType {
        &mut self.return_type
    }
}
