use crate::vm::function::DvmFunction;
use crate::vm::implementation::DvmImplementation;
use crate::vm::interface::DvmInterface;
use crate::vm::DvmContext;

pub struct DvmObjectFile {
    path: DvmPath,
    dependencies: Vec<DvmPath>,
    interfaces: Vec<DvmInterface>,
    implementations: Vec<DvmImplementation>,
    functions: Vec<DvmFunction>,
}

#[derive(Debug, Clone, PartialEq, Eq)]
pub struct DvmPath {
    base_fqn: String,
    file_name: String,
}

impl DvmPath {
    pub fn new(base_fqn: &str, file_name: &str) -> Self {
        DvmPath {
            base_fqn: base_fqn.to_string(),
            file_name: file_name.to_string(),
        }
    }
}

impl DvmObjectFile {
    pub fn new(path: DvmPath) -> Self {
        DvmObjectFile {
            path,
            dependencies: vec![],
            interfaces: vec![],
            implementations: vec![],
            functions: vec![],
        }
    }

    pub fn add_function(&mut self, function: DvmFunction) {
        self.functions.push(function);
    }

    pub fn add_dependency(&mut self, path: DvmPath) {
        self.dependencies.push(path);
    }

    pub fn load_to_context(
        self,
        context: &mut DvmContext,
        linker: &impl Fn(&DvmPath) -> DvmObjectFile,
    ) {
        if context.is_object_file_loaded(&self.path) {
            return;
        }

        // Offload our own declarations
        for interface in self.interfaces {
            context.add_interface(interface);
        }
        for implementation in self.implementations {
            context.add_implementation(implementation);
        }
        for function in self.functions {
            context.add_static_function(function);
        }

        context.mark_object_file_as_loaded(self.path);

        // Recursively load dependencies
        for dependency in &self.dependencies {
            let dependency_object_file = linker(dependency);
            dependency_object_file.load_to_context(context, linker);
        }
    }
}
